from typing import Union
import CharActor as ca
import CharMonster as cm
import entyty
from random import choice
import re

def all_equal(iterable):
    return all(x == iterable[0] for x in iterable)

def create_enemy(enemy_number, grid):
    return cm.Monster(enemy_number, grid=grid)
    # ca.create(obj_name, enemy_name, role='Fighter', race='Human', alignment='chaotic_evil', background='Soldier', grid = grid)
    # return getattr(ca.character_bank, obj_name)

def check_line_of_sight(grid, enemy: Union[entyty.GridEntity, ], player: Union[entyty.GridEntity, ]):
    enemy_cell = enemy.cell.designation
    player_cell = player.cell.designation
    enemy_vision = enemy.vision // 10
    if grid.get_distance(enemy_cell, player_cell) <= enemy_vision:
        path, cost = grid.get_path(enemy_cell, player_cell)
        directions = []
        for i, cell in enumerate(path):
            cell = grid[cell]
            next_cell = grid[path[i + 1]] if i + 1 < len(path) else None
            if next_cell is None:
                break
            direction = grid.get_direction(cell, next_cell)
            directions.append(direction)
        if all_equal(directions):
            return True
    return False

def process_enemy_actions(grid, enemy: Union[entyty.GridEntity, ], player: Union[entyty.GridEntity, ]):
    if enemy.hp <= 0:
        return 'dead'
    enemy_cell = enemy.cell.designation
    player_cell = player.cell.designation
    enemy_vision = 20
    if grid.get_distance(enemy_cell, player_cell) <= enemy_vision:
        if grid.get_distance(enemy_cell, player_cell) == 1:
            enemy.set_target(player)
            attack_result = enemy._attack_figure()
            return attack_result
        # if check_line_of_sight(grid, enemy, player):
        path, cost = grid.get_path(enemy_cell, player_cell)
        if len(path) > 1:
            enemy.move(cell=path[0])
            return 'move'
    action = choice(['move', 'wait', 'wait'])
    if action == 'move':
        direction = choice(['north', 'south', 'east', 'west'])
        enemy.move(direction=direction)
        return 'move'
    else:
        return 'wait'



from typing import Type
import random
from CharObj import Item as _Item
from grid_engine import grid as Grid, cell as Cell, grid_object as GridObject


class GridItemMetaFactory:
    """This class merges the functionality of the Goods and Armory catalogues 
    present in the CharObj module and the GridItem class present in the GridObject submodule of the grid_engine module.
    The purpose of this class is to allow the creation of items that inherit from GridItem and a class of item present 
    in the catalogues. Thereby, the items created will automatically be added to the grid. The items created with this class
    can be interacted with by characters created using the CharActor module, provided the two inhabit the same grid. 
    The class also provides a method to create a random item and a method to create an item by name. The class also 
    provides a method to check that the GRID variable has been set before any items are created.
    
    Attributes:
        GRID (type(Grid.Grid)): The grid to add the items to
        GOODS (type(Goods)): The goods catalogue
        ARMORY (type(Armory)): The armory catalogue
        CLASSES (list): A list of all the classes in the catalogue
    
    Methods:
        init_grid(grid: type(Grid.Grid)): This method sets the variable GRID and CLASSES to the grid and the classes in the catalogues respectively.
        create_item_by_class(Class: type
        create_random_item(cell = None): This method creates an instance of a random class that inherits from GridItem and a class of item present in the catalogues.
        create_item(item_name, cell = None): This method creates an instance of a class that inherits from GridItem and a class of item present in the catalogues.
        _check_grid(): This method checks that the GRID variable has been set before any items are created.
    """
    GRID = None
    GOODS = None
    ARMORY = None
    CLASSES = None

    @staticmethod
    def init_grid(grid: Type[Grid.Grid]):
        """This method sets the variable GRID and CLASSES to the grid and the classes in the catalogues respectively."""
        GridItemMetaFactory.GRID = grid
        GridItemMetaFactory.GOODS = grid.goods
        GridItemMetaFactory.ARMORY = grid.armory
        GridItemMetaFactory.CLASSES = list(grid.goods.items.values()) + list(grid.armory.items.values())
        
    @staticmethod
    def create_item_by_class(Class: Type[_Item], cell: Type[Cell.Cell] = None, no_cell: bool = None) -> Type[GridObject.GridItem]:
        """This method creates an instance of a class that inherits from GridItem and a class of item present in the catalogues.
        It then adds the instance to the appropriate catalogue and returns the instance. By subclassing GridItem, the item will
        automatically be added to the grid.
        
        Args:
            cell (type(Cell.Cell)): The cell to add the item to
            Class (type(_Item)): The class to create an instance of
        
        Returns:
            instance (type(_Item)): The instance of the class created"""
        GridItemMetaFactory._check_grid()
        grid = GridItemMetaFactory.GRID
        Goods = GridItemMetaFactory.GOODS
        Armory = GridItemMetaFactory.ARMORY
        if not no_cell:
            cell = cell if cell is not None else grid.random_cell(attr=('passable', True))
        else:
            cell = None
        if Class in GridItemMetaFactory.CLASSES:
            grid_class = type(Class.__name__, (GridObject.GridItem, Class), {}) # This creates a new class that inherits from GridItem and the class passed in as an argument
            instance = grid_class(grid=grid, cell=cell, name=Class.__name__) # This creates an instance of the new class and passes in the grid and cell arguments
        grid_instances = {}
        grid_instances |= Goods._grid_instances
        grid_instances |= Armory._grid_instances
        if grid_instances.get(instance.name) is None:
            if instance.name in Goods:
                Goods._grid_instances[instance.name] = instance
            elif instance.name in Armory:
                Armory._grid_instances[instance.name] = instance
        else:
            item_count = sum(bool(item_name.startswith(instance.name))
                         for item_name in grid_instances)
            if instance.name in Goods:
                Goods._grid_instances[f'{instance.name}{item_count+1}'] = instance                                        
            elif instance.name in Armory:
                Armory._grid_instances[f'{instance.name}{item_count+1}'] = instance
        return instance 
    
    @staticmethod
    def create_random_item(cell = None):
        """This method creates an instance of a random class that inherits from GridItem and a class of item present in the catalogues."""
        Class = random.choice(GridItemMetaFactory.CLASSES)
        return GridItemMetaFactory.create_item_by_class(Class, cell)
    
    @staticmethod
    def create_item(item_name, cell = None):
        """This method creates an instance of a class that inherits from GridItem and a class of item present in the catalogues."""
        class_name = item_name.title().replace(' ', '', len(item_name.split()) - 1)
        class_names = [cls.__name__ for cls in GridItemMetaFactory.CLASSES]
        try:
            Class = GridItemMetaFactory.CLASSES[class_names.index(class_name)]
        except ValueError as e:
            raise ValueError(f'{item_name} is not a valid item name.') from e
        return GridItemMetaFactory.create_item_by_class(cell, Class)
    
    @staticmethod
    def _check_grid():
        grid = GridItemMetaFactory.GRID if GridItemMetaFactory.GRID is not None else None
        if grid is None:
            raise ValueError('GridItemMetaFactory.GRID must be set before attempting to create any items. Use GridItemMetaFactory.init_grid(grid) to set the grid.')




from chargrid import Game, Grid, Item, ItemStack
import grid_engine as ge
import curses
from colorama import Fore
from subprocess import call
import time
from time import sleep
import numpy as np
import os, sys
import re

def clear():return call('clear')
def tick():sleep(0.1),clear()
def long_tick():sleep(1),clear()
def stop():sleep(3), clear()

r = Fore.RESET
red = Fore.LIGHTRED_EX
green = Fore.LIGHTGREEN_EX
blue = Fore.LIGHTBLUE_EX
yellow = Fore.LIGHTYELLOW_EX
magenta = Fore.LIGHTMAGENTA_EX
cyan = Fore.LIGHTCYAN_EX
white = Fore.LIGHTWHITE_EX

# sys.ps1 = f'{blue}>>> {white}'

# tick()

# print(f'{white}Preparing {blue}demo{white}...')
# stop()

# print(f'Creating {blue}game{white}...')
# stop()


# print(f'Call {blue}game{white} using `{green}game{white}` ...')
# stop()

# print(f'Identifying {blue}character{white}...')
# stop()

# print(f'Call {blue}character{white} using `{green}char{white}` ...')
# stop()

# print(f'Identifying {blue}grid{white} ...')
# stop()


# print(f'Call {blue}grid{white} using `{green}grid{white}` ...')
# stop()

# tick()

# print('All done.')
# stop()

def determine_sections(grid, char_x, char_y, max_x, max_y):
    # Determine the sections of the grid that should be drawn based on the character's initial position
    # The sections will follow a (x, y) coordinate system with the origin at the top left corner of the grid
    # The sections will be returned as a np.array of tuples
    # Each tuple will contain the top left coordinates and bottom right coordinates of the section
    # The sections will be drawn as needed based on the character's movement
    # initialize the array
    sections = np.array([((0, 0), (max_x-2, max_y-2))])
    # determine the number of sections needed
    # the number of sections will be determined by the size of the grid and the size of the terminal
    section_count_x = grid.blueprint.col_count // (max_x - 2)
    section_count_y = grid.blueprint.row_count // (max_y - 2)
    first_screen_index = 0
    if section_count_x == 1 and section_count_y == 1:   
        return sections, first_screen_index
    else:
        for y in range(section_count_y):
            for x in range(section_count_x):
                if x == 0 and y == 0:
                    continue
                x1, y1 = (x * (max_x-2), y * (max_y - 2))
                x2, y2 = (x1 + (max_x-2), y1 + (max_y - 2))
                sections = np.append(sections, [((x1, y1), (x2, y2))], axis=0)
                if char_x in range(x1, x2) and char_y in range(y1, y2):
                    first_screen_index = len(sections) - 1
        return sections, first_screen_index

# Draw functions

def draw_screen(stdscr, grid, section, enemy):
    max_y, max_x = stdscr.getmaxyx()
    section_tl, section_br = section[0], section[1]

    # draw screen
    for y in range(section_tl[1], section_br[1]):
        for x in range(section_tl[0], section_br[0]):
            win_x, win_y = (min((x - section_tl[0]), max_x - 2), min((y - section_tl[1]), max_y - 2))
            cell_x, cell_y = x, y

            cell = grid[(x, y)]
            items_in_cell = cell.entry_object['items']
            structures_in_cell = cell.entry_object['structures']
            entry_zone = cell.entry_zone

            if grid.with_terrain:
                terrain_str = cell.terrain_str
                terrain_char = cell.terrain_char
                draw_terrain(stdscr, win_x, win_y, terrain_str, terrain_char)
            if entry_zone['areas'] is not None:
                draw_structure(stdscr, win_x, win_y, entry_zone, structures_in_cell)
            if cell.passable:
                stdscr.addstr(win_y, win_x, ' ')
            else:
                stdscr.addstr(win_y, win_x, ' ', curses.color_pair(4))
            draw_item(stdscr, win_x, win_y, items_in_cell)
            draw_occupied_cells(stdscr, grid, cell_y, cell_x, win_x, win_y, enemy)

def draw_terrain(stdscr, win_x, win_y, terrain_str, terrain_char):
    if terrain_str == 'OCEAN':
        stdscr.addstr(win_y, win_x, '~', curses.color_pair(1))
    elif terrain_char == ' ':
        stdscr.addstr(win_y, win_x, ' ', curses.color_pair(0))
    elif terrain_str in ['SAND', 'SEASHELL']:
        stdscr.addstr(win_y, win_x, ' ', curses.color_pair(0))
    elif terrain_char == '^':
        stdscr.addstr(win_y, win_x, '^', curses.color_pair(2))

def draw_item(stdscr, win_x, win_y, items_in_cell):
    if items_in_cell is not None:
        if len(list(items_in_cell.keys())) > 0:
            item_key = list(items_in_cell.keys())[0]
            item = items_in_cell[item_key]
            stdscr.addstr(win_y, win_x, item.name[0].lower())

def draw_structure(stdscr, win_x, win_y, entry_zone, structs):
    if len(list(entry_zone['areas'].keys())) > 0:
        area_key = list(entry_zone['areas'].keys())[0]
        if area_key.startswith('room'):
            if structs is not None:
                if len(list(structs.keys())) > 0:
                    structure_keys = list(structs.keys())
                    for structure_key in structure_keys:
                        if structure_key == 'room':
                            stdscr.addstr(win_y, win_x, ' ', curses.color_pair(0))
                        elif structure_key == 'wooden_wall':
                            stdscr.addstr(win_y, win_x, '#', curses.color_pair(0))
                        elif structure_key == 'door':
                            stdscr.addstr(win_y, win_x, '_', curses.color_pair(0))            

def draw_action_menu(action_menu_win, selection):
    action_menu_win.clear()
    action_menu_win.box()
    action_menu_win.addstr(1, 2, 'Actions')
    action_menu_win.addstr(2, 2, f'1. View Inventory{'<' if selection == 1 else ''}')
    action_menu_win.addstr(3, 2, f'2. View Stats{'<' if selection == 2 else ''}')
    action_menu_win.addstr(4, 2, f'3. View Map{'<' if selection == 3 else ''}')
    action_menu_win.addstr(5, 2, f'4. Setup Camp{'<' if selection == 4 else ''}')
    action_menu_win.addstr(6, 2, f'5. View key bindings{'<' if selection == 5 else ''}')
    action_menu_win.addstr(7, 2, f'6. Settings{'<' if selection == 6 else ''}')
    action_menu_win.addstr(8, 2, f'7. Save Game{'<' if selection == 7 else ''}')
    action_menu_win.addstr(9, 2, f'8. Quit{'<' if selection == 8 else ''}')
    action_menu_win.refresh()
    
def draw_stats(sub_menu_win, character):
    sub_menu_win.clear()
    sub_menu_win.box()
    sub_menu_win.addstr(1, 2, f'{character.name} Level {character.level} {character._alignment.title} {character._race.title} {character._role.title}')
    sub_menu_win.addstr(3, 2, f'HP: {character.hp}')
    sub_menu_win.addstr(3, 6, f'AC: {character.armor_class}')
    sub_menu_win.addstr(5, 2, f'{character.Strength}')
    sub_menu_win.addstr(6, 2, f'{character.Dexterity}')
    sub_menu_win.addstr(7, 2, f'{character.Constitution}')
    sub_menu_win.addstr(5, 16, f'{character.Intelligence}')
    sub_menu_win.addstr(6, 16, f'{character.Wisdom}')
    sub_menu_win.addstr(7, 16, f'{character.Charisma}')
    sub_menu_win.addstr(9, 2, f'     Saving Throws')
    sub_menu_win.addstr(10, 2, f'Fort       Ref       Will')
    sub_menu_win.addstr(11, 2, f' {character.saving_throws["Fortitude"]}          {character.saving_throws["Reflex"]}          {character.saving_throws["Will"]}')
    weapon = character.inventory.equipment['MAIN_HAND']
    if weapon is not None:
        sub_menu_win.addstr(13, 2, f'Weapon: {character.inventory.equipment['MAIN_HAND'].name}')
        sub_menu_win.addstr(13, 25, 'Damage        Damage Type       Range')
        sub_menu_win.addstr(14, 2, f'Atk Bonus: +{character.Strength.modifier}')
        sub_menu_win.addstr(14, 26, f'{character.inventory.equipment["MAIN_HAND"].damage[0]}d{character.inventory.equipment['MAIN_HAND'].damage[1].value}          {character.inventory.equipment['MAIN_HAND'].damage_type}      {character.inventory.equipment["MAIN_HAND"].weapon_range}')
    else:
        sub_menu_win.addstr(13, 2, 'Unarmed')

    sub_menu_win.addstr(1, 128, 'Skills')
    for i, skill in enumerate(character.skillbook.items.keys()):
        sub_menu_win.addstr(3 + i, 128, f'{skill.replace('_', ' ')}: {character.skillbook.values()[i]}')
    sub_menu_win.refresh()

def draw_sub_menu(sub_menu_win, selected_menu, character):
    sub_menu_win.addstr(0, ((sub_menu_win.getmaxyx()[1]//2) - len(selected_menu)//2), selected_menu)
    sub_menu_win.refresh()
    
def draw_inventory(sub_menu_win, character, selected_item=1):
    max_y, max_x = sub_menu_win.getmaxyx()
    for i, item in enumerate(character.inventory, start=1):
        item = re.sub(r'(?<!\s)([A-Z])', r' \1', item)
        if i < max_y:
            sub_menu_win.addstr(i, 2, f'{'>' if selected_item == i else ''}{item}     ')    
        else:
            sub_menu_win.addstr((i) - (max_y - 2), 30 , f'{'>' if selected_item == i else ''}{item}     ')
    sub_menu_win.refresh()
    
def draw_enemy_name(stdscr, enemy):
    max_y, max_x = stdscr.getmaxyx()
    stdscr.addstr(max_y - 3, max_x - 30, f'{enemy.monster_name}')
    
def draw_time(time_win, step_count):
    time_win.clear()
    time_win.box()
    time_win.addstr(1, 2, f'Time: {get_time(step_count)}')
    time_win.refresh()
    
def draw_occupied_cells(stdscr, grid, y, x, win_x, win_y, enemy):
    cell = grid[(x, y)]
    enemy_cell = enemy.cell
    y, x = win_y, win_x

    is_cell_occupied = cell.occupied
    is_screen_occupied = stdscr.inch(y, x) == ord('@')
    
    is_enemy_dead = enemy.hp <= 0

    if is_cell_occupied and not is_screen_occupied:
        stdscr.addstr(y, x, '@')
    elif not is_cell_occupied and is_screen_occupied:
        stdscr.addstr(y, x, ' ')
        
    if cell == enemy_cell and is_enemy_dead:
        stdscr.addstr(y, x, '%')

# Echo functions

def echo_mvmnt(mvmnt_box, mvmnt):
    if mvmnt is not None:
        mvmnt_box.box()
        mvmnt_box.clear()
        mvmnt_box.addstr(1, 2, f'{mvmnt}')
        mvmnt_box.refresh()
        
def echo_pickup(pickup_box, item_name):
    if item_name is not None:
        pickup_box.box()
        pickup_box.clear()
        pickup_box.addstr(1, 2, f'Picked up{item_name}')
        pickup_box.refresh()
        
def echo_damage(damage_win, attack):
    if attack is not None:
        damage_win.box()
        damage_win.clear()
        damage_win.addstr(1, 2, f'{attack}')
        damage_win.refresh()
        
def echo_attack(attack_win, attack):
    if attack is not None:
        attack_win.box()
        attack_win.clear()
        attack_win.addstr(1, 2, f'{attack}         ')
        attack_win.refresh()

def echo_results(mvmnt, atk, previous_atk, item_name, enemy_action, damage_win, atk_win, mvmnt_box, pickup_box, time_win, game_over_win, step_count):
    if mvmnt is not None:
        echo_mvmnt(mvmnt_box, mvmnt)
    if atk is not None:
        echo_attack(atk_win, atk)
    else:
        echo_attack(atk_win, previous_atk)    
    if item_name is not None:
        echo_pickup(pickup_box, item_name)    
    if enemy_action is not None and enemy_action not in ['move', 'wait']:
        echo_damage(damage_win, enemy_action)
    draw_time(time_win, step_count)

# Input processing functions
   
def process_input(key, stdscr, character, section, grid, action_menu_win, sub_menu_win, build_win):
    if key == 'x':
        action_menu(key, stdscr, action_menu_win, sub_menu_win, character, grid)
        result = (None, None)
    elif key in '123456789':
        result = process_movement_input(key, character)
    elif key == 'a':
        result = process_attack_input(key, stdscr, character, section, grid)
    elif key == 'z':
        result = process_pickup_input(key, character, grid)
        if result is None:
            result = (None, None)
    elif key == 'b':
        process_build_input(key, stdscr, build_win, character, section, grid)
        result = (None, None)
    else:
        result = (None, None)
    return result

def process_movement_input(key, char):
    mvmnt = None
    if key == '1':
        mvmnt = char.move('south_west')
    elif key == '2':
        mvmnt = char.move('south')
    elif key == '3':
        mvmnt = char.move('south_east')
    elif key == '4':
        mvmnt = char.move('west')
    elif key == '5':
        char.end_turn()
    elif key == '6':
        mvmnt = char.move('east')
    elif key == '9':
        mvmnt = char.move('north_east')
    elif key == '8':
        mvmnt = char.move('north')
    elif key == '7':
        mvmnt = char.move('north_west')
    return ('mvmnt', mvmnt)

def process_attack_input(key, stdscr, character, section, grid):
    return ('atk', player_attack(stdscr, character, section, grid))

def process_pickup_input(key, character, grid):
    for cell in character.cell.adjacent:
        if cell is not None:
            if grid[cell].entry_object['items'] is not None:
                if len(list(grid[cell].entry_object['items'].keys())) > 0:
                    item_key = list(grid[cell].entry_object['items'].keys())[0]
                    item = grid[cell].entry_object['items'][item_key]
                    item_name = item.name
                    character.inventory.pick_up(item)
                    grid[cell].entry_object['items'] = {}
                    return ('item_name', re.sub(r'(?<!\s)([A-Z])', r' \1', item_name))
    return None

def process_build_input(key, stdscr, build_win, character, section, grid):
    if key == 'b':
        build_object = None
        build_cell = None
        selected_cell = None
        selection = 1
        char_cell = character.cell
        relative_index = 0
        while build_object is None:
            build_win.clear()
            build_win.box()
            build_win.addstr(1, 2, 'Build')
            build_win.addstr(2, 2, f'1. Wooden Wall{"<" if selection == 1 else ""}')
            build_win.addstr(3, 2, f'2. Wooden Door{"<" if selection == 2 else ""}')
            build_win.refresh()
            key = stdscr.getkey()
            if key == "KEY_DOWN":
                selection += 1
                if selection > 2:
                    selection = 1
            elif key == "KEY_UP":
                selection -= 1
                if selection < 1:
                    selection = 2
            elif key == "b":
                if selection == 1:
                    build_object = 'wooden_wall'
                elif selection == 2:
                    build_object = 'door'
                break
        build_win.clear()
        build_cell = choose_cell(stdscr, grid, character, section, 'b')
        if build_object == 'wooden_wall':
            ge.grid_object.build_wall(grid, build_cell)
        elif build_object == 'door':
            ge.grid_object.build_door(grid, build_cell)
        curses.curs_set(0)
        return
        
def player_attack(stdscr, character, section, grid):
    attack_cell = choose_cell(stdscr, grid, character, section, 'a')
    if attack_cell.occupant is not None:
        character.target = attack_cell.occupant.parent
        if character.target is not None:
            return character.attack()
    return None

def action_menu(key, stdscr, action_menu_win, sub_menu_win, character, grid):
        sub = interact_action(stdscr, action_menu_win)
        sub_menu = get_sub_menu(sub_menu_win, sub)
        draw_sub_menu(sub_menu_win, sub_menu, character)
        if sub_menu == 'Inventory':
            interact_inventory(stdscr, sub_menu_win, character)
        elif sub_menu == 'Stats':
            draw_stats(sub_menu_win, character)
            key = stdscr.getkey()

def get_relative_coordinates(stdscr, section, target_cell):
    max_y, max_x = stdscr.getmaxyx()
    section_tl, section_br = section[0], section[1]
    cell_x, cell_y = target_cell.coordinates
    
    return (min(1 + (cell_x - section_tl[0]), max_x - 2), min(1 + (cell_y - section_tl[1]), max_y - 2))    
    
def interact_inventory(stdscr, sub_menu_win, character):
    selected_item = 1
    inspect = False
    leave = False
    while not inspect and not leave:
        draw_inventory(sub_menu_win, character, selected_item)
        key = stdscr.getkey()
        if key == 'KEY_DOWN':
            selected_item += 1
            if selected_item > len(character.inventory.items):
                selected_item = 1
        elif key == 'KEY_UP':
            selected_item -= 1
            if selected_item < 1:
                selected_item = len(character.inventory.items)
        elif key == 'i':
            inspect = True
        elif key == 'x':
            leave = True
    if inspect:
        sub_menu_win.clear()
        sub_menu_win.box()
        if isinstance(character.inventory.items[selected_item - 1], ItemStack):
            pass
        sub_menu_win.addstr(2, 2, character.inventory.items[selected_item - 1].description)
        sub_menu_win.refresh()
        stdscr.getkey()
    elif leave:
        sub_menu_win.clear()
        sub_menu_win.refresh()
        
def interact_action(stdscr, action_menu_win):
    selection = 1
    sub = None
    while sub == None:
        draw_action_menu(action_menu_win, selection)
        key = stdscr.getkey()
        if key == 'KEY_DOWN':
            selection += 1
            if selection > 8:
                selection = 1
        elif key == 'KEY_UP':
            selection -= 1
            if selection < 1:
                selection = 8
        elif key == 'x':
            sub = selection
            selection = None
    return sub
                    
def choose_cell(stdscr, grid, character, section, confirmation_key):
        char_cell = character.cell
        chosen_cell = None
        selected_cell = grid[char_cell.adjacent[0]]
        y, x = stdscr.getyx()
        curses.curs_set(1)
        relative_index = 0
        while chosen_cell is None:
            selected_cell = grid[char_cell.adjacent[relative_index]]
            relative_coords = get_relative_coordinates(stdscr, section, selected_cell)
            stdscr.move(relative_coords[1]-1, relative_coords[0]-1)
            key = stdscr.getkey()
            if key == "KEY_DOWN":
                if relative_index == 0:
                    relative_index = 7
                elif relative_index in [2, 3]:
                    relative_index += 1
                elif relative_index == 7:
                    relative_index = 6
                elif relative_index == 1:
                    relative_index = 5                
            elif key == "KEY_UP":
                if relative_index == 7:
                    relative_index = 0
                elif relative_index in [3, 4]:
                    relative_index -= 1
                elif relative_index == 6:
                    relative_index = 7
                elif relative_index == 5:
                    relative_index = 1
            elif key == "KEY_LEFT":
                if relative_index in [1, 2]:
                    relative_index -= 1
                elif relative_index in [4, 5]:
                    relative_index += 1
                elif relative_index == 3:
                    relative_index = 7
            elif key == "KEY_RIGHT":
                if relative_index in [0, 1]:
                    relative_index += 1
                elif relative_index in [5, 6]:
                    relative_index -= 1
                elif relative_index == 7:
                    relative_index = 3
            if key == confirmation_key:
                chosen_cell = selected_cell
                break
        stdscr.move(y, x)
        curses.curs_set(0)
        return chosen_cell
                    
def get_sub_menu(sub_menu_win, selection):
    sub_menu_win.clear()
    sub_menu_win.box()
    sub_menus = [
        'Inventory',
        'Stats',
        'Map',
        'Camp',
        'Key Bindings',
        'Settings',
        'Save Game',
        'Quit'
]
    return sub_menus[selection - 1]

def get_time(step_count):
    seconds = (step_count * 10) % 60
    minutes = ((step_count * 10) // 60) % 60
    hours = (((step_count * 10) // 60) // 60) % 24
    days = ((((step_count * 10) // 60) // 60) // 24)
    return f'{days}d {hours}h {minutes}m {seconds}s'
    
def main(stdscr): # type: curses._CursesWindow
    curses.curs_set(0)
    curses.init_pair(1, curses.COLOR_WHITE, curses.COLOR_BLUE)
    curses.init_color(curses.COLOR_GREEN, 74, 255, 128)
    curses.init_pair(2, curses.COLOR_WHITE, curses.COLOR_GREEN)
    curses.init_color(curses.COLOR_YELLOW, 255, 130, 145)
    curses.init_pair(3, curses.COLOR_YELLOW, curses.COLOR_YELLOW)
    curses.init_color(8, 133, 133, 133)
    curses.init_pair(4, 8, 8)

    stdscr.keypad(True)    
    
    max_y, max_x = stdscr.getmaxyx()        
    print(f'Terminal size: Rows = {max_y}, Columns = {max_x}')

    grid = None
    mvmnt_box = curses.newwin(3, 22, max_y - 5, max_x - 55)
    pickup_box = curses.newwin(3, 30, max_y - 10, max_x - 55)
    action_menu_win = curses.newwin(12, 30, max_y // 4, max_x // 4)
    sub_menu_win = curses.newwin(max_y - 10, max_x - 20, 5, 10)
    time_win = curses.newwin(3, 30, 3, 3)
    build_win = curses.newwin(12, 30, max_y // 4, max_x // 4)
    damage_win = curses.newwin(3, 30, max_y - 15, max_x - 55)
    game_over_win = curses.newwin(max_y // 3, max_x // 3, max_y // 3, max_x // 3)
    atk_win = curses.newwin(3, 30, max_y - 20, max_x - 55)
    step_count = 0
    game_time = 0
    
        
    while grid is None:
        sys.stdout = open(os.devnull, 'w')
        # big grid
        # dimensions = (max_x - 2) * 5, (max_y - 2) * 5
        # small grid
        dimensions = max_x - 2, max_y - 2
        
        game = Game(map_type='dungeon', dimensions=dimensions)

        grid: Grid.Grid = game.grid
    
    sys.stdout = sys.__stdout__

    char = game.player
    
    monster = game.enemy
    
    char_x, char_y = char.cell.coordinates
    
    sections, first_screen_index = determine_sections(grid, char_x, char_y, max_x, max_y)
    
    current_section = sections[first_screen_index]
    
    draw_screen(stdscr, grid, current_section, monster)
    
    mvmnt = None
    atk = None
    previous_atk = None
    selection = None
    item_name = None
    enemy_action = None
    previous_enemy_action = None
    result = (None, None)
        
    while True:
        for i, section in enumerate(sections):
            if char_x in range(section[0][0], section[1][0]) and char_y in range(section[0][1], section[1][1]):
                current_section = section
        char_x, char_y = char.cell.coordinates
        
        key = stdscr.getkey()
        if key == 'q':
            break
        else:
            result = process_input(key, stdscr, char, current_section, grid, action_menu_win, sub_menu_win, build_win)
        
        if result[0] == 'mvmnt':
            mvmnt = result[1]
        if result[0] == 'atk':
            atk = result[1]
            previous_atk = atk
        else:
            previous_atk = atk
            atk = None
        if result[0] == 'item_name':
            item_name = result[1]
        
        if any([mvmnt, atk]):
            step_count += 1

        if game.enemy.hp > 0:
            enemy_action = game.process_enemy_actions()
        else:
            enemy_action = None
        previous_enemy_action = enemy_action if enemy_action is not None and enemy_action not in ['move', 'wait'] else previous_enemy_action

        draw_screen(stdscr, grid, current_section, monster)
        stdscr.refresh()
        echo_results(mvmnt, atk, previous_atk, item_name, enemy_action, damage_win, atk_win, mvmnt_box, pickup_box, time_win, game_over_win, step_count)
        # echo_mvmnt(mvmnt_box, mvmnt)
        # echo_pickup(pickup_box, item_name)
        # echo_damage(damage_win, enemy_action)
        # echo_attack(atk_win, previous_atk)    
        # draw_time(time_win, step_count)
        draw_enemy_name(stdscr, game.enemy)
        if char.hp <= 0:
            game_over_win.box()
            game_over_win.addstr((max_y // 3) // 2, ((max_x // 3) // 2) - 4, 'Game Over')
            game_over_win.refresh()
            stdscr.getkey()
            break


curses.wrapper(main)


import random
from typing import Any, Union, Type
from CharActor import create, character_bank
from CharActor._charactor.actor._actor.base_actor import BaseCharacter as _BaseCharacter
from CharTask import task_list as _task_list
from CharObj import Goods, Armory, Item, ItemStack
from grid_engine import grid as Grid, grid_object as GridObject
from .meta import *
from . import enemy
from .enemy import *




# Let's create a class to hold our game state
# The game state will include the grid and the characters
# We'll also attach the Catalogues, `Goods` and `Armory` to the grid object
# so that the characters can access them easily.

import random
import json
import grid_engine as ge

class Dungeon:
    def __init__(self, width, height, max_rooms, min_room_size, max_room_size):
        self.width = width
        self.height = height
        self.max_rooms = max_rooms
        self.min_room_size = min_room_size
        self.max_room_size = max_room_size
        self.grid = ge.grid.Grid(cell_size=1, dimensions=(width, height), with_terrain=False)
        for cell in self.grid:
            self.grid[cell].passable = False
        self.rooms = []
        self.room_groups = {}
        
    def create_structure(self, x, y, structure):
        self.grid[x, y].passable = True
        return ge.grid_object.GridStructure(self.grid, structure, self.grid[x, y], passable=True)
        
    def create_room(self, x, y, w, h):
        room_count = len(self.rooms)
        room_center = (x + w // 2, y + h // 2)
        room = ge.grid_object.GridZone(self.grid, f'room{room_count+1}', 'area', room_center)
        for j in range(x, x + w):
            for i in range(y, y + h):
                structure = self.create_structure(j, i, 'room')
                cell = self.grid[j, i]
                cell.add_object(structure)
                cell.add_zone(room)
                room._add_cell(cell)
        self.room_groups[room.name] = room
                
    def is_room_valid(self, x, y, w, h):
        if x + w >= self.width or y + h >= self.height:
            return False
        for i in range(x - 1, x + w + 1):
            for j in range(y - 1, y + h + 1):
                if self.grid[i, j].entry_object['structures'] is not None:
                    return False
        return True

    def place_rooms(self):
        for _ in range(self.max_rooms):
            w = random.randint(self.min_room_size, self.max_room_size)
            h = random.randint(self.min_room_size, self.max_room_size)
            x = random.randint(0, self.width - w - 1)
            y = random.randint(0, self.height - h - 1)

            if self.is_room_valid(x, y, w, h):
                self.create_room(x, y, w, h)
                self.rooms.append((x, y, w, h))

    def connect_rooms(self):
        for i in range(len(self.rooms) - 1):
            x1, y1, w1, h1 = self.rooms[i]
            x2, y2, w2, h2 = self.rooms[i + 1]
            cx1, cy1 = x1 + w1 // 2, y1 + h1 // 2
            cx2, cy2 = x2 + w2 // 2, y2 + h2 // 2

            if random.choice([True, False]):
                self.create_h_corridor(cx1, cx2, cy1)
                self.create_v_corridor(cy1, cy2, cx2)
            else:
                self.create_v_corridor(cy1, cy2, cx1)
                self.create_h_corridor(cx1, cx2, cy2)

    def create_h_corridor(self, x1, x2, y):
        for x in range(min(x1, x2), max(x1, x2) + 1):
            if self.grid[x, y].entry_object['structures'] is None:
                structure = self.create_structure(x, y, 'corridor')
                self.grid[x, y].add_object(structure)

    def create_v_corridor(self, y1, y2, x):
        for y in range(min(y1, y2), max(y1, y2) + 1):
            if self.grid[x, y].entry_object['structures'] is None:
                structure = self.create_structure(x, y, 'corridor')
                self.grid[x, y].add_object(structure)

                    
    def save_to_json(self, filename):
        data = {
            'width': self.width,
            'height': self.height,
            'max_rooms': self.max_rooms,
            'min_room_size': self.min_room_size,
            'max_room_size': self.max_room_size,
            'grid': self.grid,
            'rooms': self.rooms
        }
        with open(filename, 'w') as f:
            json.dump(data, f, indent=4)

    def load_from_json(self, filename):
        with open(filename, 'r') as f:
            data = json.load(f)
        self.width = data['width']
        self.height = data['height']
        self.max_rooms = data['max_rooms']
        self.min_room_size = data['min_room_size']
        self.max_room_size = data['max_room_size']
        self.grid = data['grid']
        self.rooms = data['rooms']

    def print_dungeon(self):
        for y in range(self.height):
            for x in range(self.width):
                if self.grid[x, y].entry_object['structures'] is not None:
                    print(' ', end='')
                else:
                    print('#', end='')
            print()


# Parameters for the dungeon

class Game:
    _grid = None
    _characters = None
    _player = None
    _enemy = None
    _task_list = None

    @property
    def grid(self) -> Type[Grid.Grid]:
        return self._grid
    
    @grid.setter
    def grid(self, grid: Type[Grid.Grid]) -> None:
        self._grid = grid
        
    @property
    def characters(self) -> list:
        if self._characters is None:
            self._characters = []
        return self._characters
    
    @characters.setter
    def characters(self, characters: list) -> None:
        if self._characters is None:
            self._characters = []
        self._chaaracters = characters
        
    @property
    def player(self) -> _BaseCharacter:
        return self._player        

    @player.setter
    def player(self, player: _BaseCharacter):
        self._player = player

    @property
    def enemy(self) -> _BaseCharacter:
        return self._enemy
    
    @enemy.setter
    def enemy(self, enemy: _BaseCharacter):
        self._enemy = enemy

    @property
    def task_list(self) -> list:
        return self._task_list
    
    @task_list.setter
    def task_list(self, task_list: list) -> None:
        self._task_list = task_list
        
    def __init__(self, map_type: str = 'land', dimensions: tuple = (540, 460)):
        print('Generating grid...')
        if map_type == 'land':
            self.grid = ge.grid.Grid(cell_size=2, dimensions=dimensions, with_terrain=False)
        elif map_type == 'dungeon':
            self.dungeon = Dungeon(*dimensions, 20, 10, 20)    
            self.dungeon.place_rooms()
            self.dungeon.connect_rooms()
            self.grid = self.dungeon.grid
        print('Generating characters...')
        self.character_count = 0
        self.add_character(self.create_random_character())
        self.player = self.characters[0]
        self.enemy = create_enemy(random.randint(0, 4), self.grid)
        print('Adding catalogues...')
        self.grid.goods = Goods
        self.grid.armory = Armory
        print('Adding items...')
        self.item_factory = GridItemMetaFactory
        self.item_factory.init_grid(self.grid)
        self.add_random_items(10)
        self.task_list = _task_list
        self.prepare_player()
        print('Done!')
        
    @property
    def items_on_grid(self) -> list:
        class GridItems(dict):
            def __init__(self):
                dict.__init__(self)
                
            def __getitem__(self, key):
                key = key.replace(' ', '').lower()
                return dict.__getitem__(self, key)
            
            def __setitem__(self, key, value):
                key = key.replace(' ', '').lower()
                return dict.__setitem__(self, key, value)
            
            def __ior__(self, other):
                for key in other:
                    self[key.lower().replace(' ', '')] = other[key]
                return self
                            
        grid_items = GridItems() 
        goods = self.grid.goods._grid_instances
        armory = self.grid.armory._grid_instances
        grid_items |= armory
        grid_items |= goods
        
        return grid_items
    
    @property
    def item_list(self) -> list:
        return list(self.grid.goods.general_manifest)+list(self.grid.goods.trade_manifest)+list(self.grid.armory.weapons_manifest)+list(self.grid.armory.armor_manifest)
    
    def pickup_item(self, item: Any, character: Any) -> None:
        character.pickup(item)
        item.cell.entry_object['items'] = {}
        item.cell = None
    
    def add_character(self, character) -> None:
        self.character_count += 1
        self.characters.append(character)
        
    def create_random_character(self):
        create(grid=self.grid)
        return getattr(character_bank, f'char{self.character_count+1}')
    
    def add_item(self, item_name: str = None, cell: Union[str, object] = None) -> None:
        if item_name is None and cell is None:
            GridItemMetaFactory.create_random_item()
        elif item_name is not None and cell is None:
            GridItemMetaFactory.create_item_by_class(item_name)
        elif item_name is None:
            GridItemMetaFactory.create_random_item(cell=cell)
        else:
            GridItemMetaFactory.create_item(item_name, cell)
            
        
    def add_random_item(self, cell = None):
        self.add_item(cell=cell)
        
    def add_random_items(self, n: int):
        for _ in range(n):
            self.add_random_item()

    def give_player_item(self, item_name: str):
        self.player.inventory.add_item(item_name)
        
    def prepare_player(self):
        for _ in range(10):
            self.player.inventory.add_item('OakLog')
            self.player.inventory.add_item('IronOre')
        self.player.inventory.add_item('Tongs')
        self.player.inventory.add_item('Axe')
        
    def process_enemy_actions(self):
        return process_enemy_actions(self.grid, self.enemy, player=self.player) 



from chargrid.meta import *
import CharActor as ca
import grid_engine as ge

# Let's define a function to load some items onto the grid.
def load_grid_items(grid):
    for _ in range(10):
        GridItemMetaFactory.create_random_item(grid)

# Let's declare our grid and attach the Catalogues, `Goods` and `Armory` to it.
grid = ge.Grid.Grid(cell_size=5, dimensions=(1920, 1080), noise_scale=511, noise_octaves=83, noise_roughness=0.3)
setattr(grid, 'goods', Goods)
setattr(grid, 'armory', Armory)

# 
# Let's render the grid and see what it looks like.
dimensions = grid.blueprint.grid_dimensions
cdata = ge._grid.extract_cell_data(grid)
cell_size = grid.cell_size
grid_id = grid.grid_id

del grid

ge._utility.generate_images(dimensions, cdata, cell_size, grid_id, True)




from setuptools import setup, find_packages

setup(
    name='chargrid',
    version='0.0.3',
    description='HackNet-style grid-based game engine framework.',
    author='James Evans',
    author_email='joesaysahoy@gmail.com',
    url='https://github.com/primal-coder/chargrid',
    packages=find_packages(),
    install_requires=['gridengine_framework', 'CharActor', 'CharMonster', 'CharObj', 'dicepy', 'entyty'],
    keywords='gridengine_framework CharActor grid character rpg game engine framework chargrid cell grid-based cell-based dnd d&d dungeons and dragons dungeon dragons'    
)